# Adapted from mlx-lm

import mlx.nn as nn
import mlx.core as mx
from mlx.utils import tree_flatten, tree_unflatten, tree_map
from typing import List, Optional, Any, Dict, Union, Tuple, NamedTuple

class _BaseCache:
    @property
    def state(self):
        return []

    @state.setter
    def state(self, v):
        if v is not None and v:
            raise ValueError("This cache has no state but a state was set.")

    @property
    def meta_state(self):
        return ""

    @meta_state.setter
    def meta_state(self, v):
        if v is not None and v:
            raise ValueError("This cache has no meta_state but a meta_state was set.")

    def is_trimmable(self):
        return False


class QuantizedKVCache(_BaseCache):
    def __init__(self, group_size: int = 64, bits: int = 8):
        self.keys = None
        self.values = None
        self.offset = 0
        self.step = 256
        self.group_size = group_size
        self.bits = bits

    def update_and_fetch(self, keys, values):
        B, n_kv_heads, num_steps, k_head_dim = keys.shape
        v_head_dim = values.shape[-1]
        prev = self.offset

        if self.keys is None or (prev + num_steps) > self.keys[0].shape[-2]:
            el_per_int = 8 * mx.uint32.size // self.bits
            new_steps = (self.step + num_steps - 1) // self.step * self.step
            shape = (B, n_kv_heads, new_steps)

            def init_quant(dim):
                return (
                    mx.zeros((*shape, dim // el_per_int), dtype=mx.uint32),
                    mx.zeros((*shape, dim // self.group_size), dtype=keys.dtype),
                    mx.zeros((*shape, dim // self.group_size), dtype=keys.dtype),
                )

            def expand_quant(x):
                new_x = mx.zeros((*shape, x.shape[-1]), dtype=x.dtype)
                return mx.concatenate([x, new_x], axis=-2)

            if self.keys is not None:
                if prev % self.step != 0:
                    self.keys, self.values = tree_map(
                        lambda x: x[..., :prev, :], (self.keys, self.values)
                    )

                self.keys, self.values = tree_map(
                    expand_quant, (self.keys, self.values)
                )
            else:
                self.keys, self.values = init_quant(k_head_dim), init_quant(v_head_dim)

        self.offset += num_steps

        keys = mx.quantize(keys, group_size=self.group_size, bits=self.bits)
        values = mx.quantize(values, group_size=self.group_size, bits=self.bits)
        for i in range(len(self.keys)):
            self.keys[i][..., prev : self.offset, :] = keys[i]
            self.values[i][..., prev : self.offset, :] = values[i]

        return tree_map(lambda x: x[..., : self.offset, :], (self.keys, self.values))

    @property
    def state(self):
        if self.offset == self.keys[0].shape[2]:
            return self.keys, self.values
        else:
            return tree_map(
                lambda x: x[..., : self.offset, :], (self.keys, self.values)
            )

    @state.setter
    def state(self, v):
        self.keys, self.values = v

    @property
    def meta_state(self):
        return tuple(map(str, (self.step, self.offset, self.group_size, self.bits)))

    @meta_state.setter
    def meta_state(self, v):
        self.step, self.offset, self.group_size, self.bits = map(int, v)

    def is_trimmable(self):
        return True

    def trim(self, n):
        n = min(self.offset, n)
        self.offset -= n
        return n


class KVCache(_BaseCache):
    def __init__(self):
        self.keys = None
        self.values = None
        self.offset = 0
        self.step = 256

    def update_and_fetch(self, keys, values):
        prev = self.offset
        if self.keys is None or (prev + keys.shape[2]) > self.keys.shape[2]:
            B, n_kv_heads, _, k_head_dim = keys.shape
            v_head_dim = values.shape[3]
            n_steps = (self.step + keys.shape[2] - 1) // self.step
            k_shape = (B, n_kv_heads, n_steps * self.step, k_head_dim)
            v_shape = (B, n_kv_heads, n_steps * self.step, v_head_dim)
            new_k = mx.zeros(k_shape, keys.dtype)
            new_v = mx.zeros(v_shape, values.dtype)
            if self.keys is not None:
                if prev % self.step != 0:
                    self.keys = self.keys[..., :prev, :]
                    self.values = self.values[..., :prev, :]
                self.keys = mx.concatenate([self.keys, new_k], axis=2)
                self.values = mx.concatenate([self.values, new_v], axis=2)
            else:
                self.keys, self.values = new_k, new_v

        self.offset += keys.shape[2]
        self.keys[..., prev : self.offset, :] = keys
        self.values[..., prev : self.offset, :] = values
        return self.keys[..., : self.offset, :], self.values[..., : self.offset, :]

    @property
    def state(self):
        if self.offset == self.keys.shape[2]:
            return self.keys, self.values
        else:
            return (
                self.keys[..., : self.offset, :],
                self.values[..., : self.offset, :],
            )

    @state.setter
    def state(self, v):
        self.keys, self.values = v
        self.offset = self.keys.shape[2]

    def is_trimmable(self):
        return True

    def trim(self, n):
        n = min(self.offset, n)
        self.offset -= n
        return n

    def to_quantized(self, group_size: int = 64, bits: int = 4) -> QuantizedKVCache:
        quant_cache = QuantizedKVCache(group_size=group_size, bits=bits)
        quant_cache.offset = self.offset
        if self.keys is not None:
            quant_cache.keys = mx.quantize(self.keys, group_size=group_size, bits=bits)
            quant_cache.values = mx.quantize(
                self.values, group_size=group_size, bits=bits
            )
        return quant_cache


class CacheHistory:
    def __init__(self,
            cache: List[Union[KVCache, QuantizedKVCache]],
            token_ids: List[List[int]]
        ) -> None:
        self.cache: List[Union[KVCache, QuantizedKVCache]] = cache
        self.token_ids: List[List[int]] = token_ids


def make_empty_cache(model: nn.Module) -> List[Union[KVCache, QuantizedKVCache]]:
    """
    Constructs an empty cache for use during generation.

    This function creates an empty cache for a given language model. If the model
    has a ``make_cache`` method, it defers the cache construction to the model.
    Otherwise, it creates a default KV cache.
    Args:
        model (nn.Module): The language model.
    Returns:
        List[Union[KVCache, QuantizedKVCache]]: The constructed empty cache.
    """
    num_layers = len(model.layers)
    return [KVCache() for _ in range(num_layers)]
    
def create_cache_dict(cache: List[Union[KVCache, QuantizedKVCache]]) -> Dict[str, mx.array]:
    """Create a dictionary of prompt cache for saving or further manipuation. 

    Args:
        cache (List[Union[KVCache, RotatingKVCache]]): Prompt cache.

    Returns:
        Dict[str, mx.array]: A dictionary of prompt cache for saving or further manipuation. 
    """
    offset = cache[0].offset
    if offset != 0:
        cache_dict = dict(tree_flatten([c.state for c in cache]))
        return cache_dict
    
def save_cache(cache: List[Union[KVCache, QuantizedKVCache]], file: str, metadata: Optional[Dict[str, str]] = None) -> None:
    """Save cache as a safetensors file.

    This function saves a given cache as a safetensors file. The cache is first
    converted to a dictionary of tensors using the `create_cache_dict` function.
    The resulting dictionary is then saved to the specified file using the
    `safetensors.save_file` function. If metadata is provided, it is included in
    the saved file.

    Args:
        cache (List[Union[KVCache, QuantizedKVCache]]): The cache to save.
        file (str): The file path to save the cache to.
        metadata (Optional[Dict[str, str]], optional): Metadata to include in the saved file. Defaults to None.
    """
    cache_dict = create_cache_dict(cache=cache)
    if cache_dict:
        mx.save_safetensors(file=file, arrays=cache_dict, metadata=metadata)
    del cache_dict
    mx.metal.clear_cache()

def load_cache(file: str) -> Tuple[List[Union[KVCache, QuantizedKVCache]], Dict[str, str]]:
    """Load a cache file.

    This function loads a cache file saved as a safetensors file. It returns the loaded cache
    and its metadata.

    Args:
        file (str): The path to the safetensors file containing the cache.

    Returns:
        Tuple[List[Union[KVCache, QuantizedKVCache]], Dict[str, str]]: A tuple containing the loaded cache
            and its metadata.
    """
    cache_dict, metadata = mx.load(file, return_metadata=True)
    cache_list = tree_unflatten(list(cache_dict.items()))
    cache = []
    for i, (key, value) in enumerate(cache_list):
        cache.append(KVCache())
        cache[i].update_and_fetch(key, value)
    mx.eval([c.state for c in cache])
    del cache_dict
    mx.metal.clear_cache()
    return cache, metadata

def split_cache(cache: List[Union[KVCache, QuantizedKVCache]]) -> List[List[KVCache]]:
    """Split the cache into multiple smaller caches. Each smaller cache will contain a subset of the original cache's data.

    Args:
        cache (List[Union[KVCache, QuantizedKVCache]]): The cache to be split.

    Returns:
        List[List[KVCache]]: A list of smaller caches.
    """
    num_split = cache[0].state[0].shape[0]
    caches = []
    def make_cache(n):
        ec = []
        for j, c in enumerate(cache):
            ec.append(KVCache())
            ec[j].update_and_fetch(c.state[0][n:n + 1], c.state[1][n:n + 1])
            mx.eval(ec[j].state)
        return ec
    for i in range(num_split):
        caches.append(make_cache(i))
    del cache
    mx.metal.clear_cache()
    return caches

def trim_cache(cache: List[KVCache], offset: int, trim_suffix: int = 0) -> List[KVCache]:
    """
    Trim the cache by removing entries before the specified offset and optionally
    removing entries after the specified trim_suffix.

    Args:
        cache (List[KVCache]): The cache to be trimmed.
        offset (int): The offset from which to start trimming.
        trim_suffix (int, optional): The number of entries after the offset to keep. Defaults to 0.

    Returns:
        List[KVCache]: The trimmed cache.
    """
    new_cache = []
    for j, c in enumerate(cache):
        new_cache.append(KVCache())
        if trim_suffix == 0:
            new_cache[j].update_and_fetch(c.state[0][:, :, offset:, :], c.state[1][:, :, offset:, :])
        else:
            new_cache[j].update_and_fetch(c.state[0][:, :, offset:-trim_suffix, :], c.state[1][:, :, offset:-trim_suffix, :])
        mx.eval(new_cache[j].state)
    del cache
    mx.metal.clear_cache()
    return new_cache


def select_from_cache(
        cache: List[Union[KVCache, QuantizedKVCache]],
        prompt_index: int = 0,
        select_len: Optional[int] = None,
        start_from: int = 0
    ) -> List[Tuple[mx.array, mx.array]]:
    """Get the kv cache tensors of a specific sequence in a cache with multiple prompt cache sequences.

    Args:
        cache (List[Union[KVCache, QuantizedKVCache]]): Cache object.
        prompt_index (int, optional): The index of the sequence to get. Defaults to 0.
        select_len (Optional[int], optional): Length of the seqence to get. Defaults to None.
        start_from (int, optional): Starting index of the sequence. Defaults to 0.

    Returns:
        List[Tuple[mx.array, mx.array]]: List of keys and values tuples for updating the new cache.
    """
    num_prompts, num_kv_heads, prompt_len, head_dim = cache[0].state[0].shape
    max_index = select_len + start_from if select_len else prompt_len
    if prompt_index >= num_prompts:
        raise ValueError('Prompt index out of index.')
    if max_index > prompt_len:
        raise ValueError(f'Select length out of index.')
    cache_list = []
    for c in cache:
        k, v  = c.state
        cache_list.append((k[prompt_index:prompt_index + 1, :, start_from:max_index, :], v[prompt_index:prompt_index + 1, :, start_from:max_index, :]))
    return cache_list

def find_max_prefix_num(new: List[int], baseline: List[int]) -> int:
    """Helper function to find the maximum number of tokens shared in the prefix of two prompts.

    Args:
        new (List[int]): First prompt token ids.
        baseline (List[int]): Second prompt token ids.

    Returns:
        int: The maximum number of tokens shared in the prefix of two prompts.
    """
    from itertools import takewhile
    return len(list(takewhile(lambda x: x[0] == x[1], zip(new, baseline))))



import os
from mlx_lm.utils import get_model_path, load_config, load_model, convert
from .tokenizer_utils import load_tokenizer, TokenizerWrapper
from logging import Logger
import mlx.nn as nn
from typing import Literal, Optional, Dict, Any, Tuple, NamedTuple
from pathlib import Path
os.environ['TOKENIZERS_PARALLELISM'] = 'true'

PACKAGE_NAME = 'mlx_textgen'

package_cache_dir = os.path.join(os.path.expanduser('~'), '.cache', PACKAGE_NAME)
os.makedirs(package_cache_dir, exist_ok=True)

def clear_prompt_cache() -> None:
    prompt_cache_dir = os.path.join(package_cache_dir, 'prompt_cache')
    if os.path.exists(prompt_cache_dir):
        from shutil import rmtree
        rmtree(prompt_cache_dir)

def get_model_name(model_id_or_path: str) -> str:
    """Getting a clean name for the model.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.

    Returns:
        str: A clean name for the model.
    """
    base_name = model_id_or_path.split('/')[-1]
    base_name = base_name.split('\\')[-1]
    base_name = base_name.lower().replace('_', '-').strip()
    return base_name

def make_model_exist(
        model_id_or_path: str,
        quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16',
        revision: Optional[str] = None,
        adapter_path: Optional[str] = None
    ) -> Path:
    """Check if the the model exists locally, if not, it will attempt to download the model and convert it to the appropriate quant.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.
        quant (Literal[&#39;fp16&#39;, &#39;q8&#39;, &#39;q4&#39;, &#39;q2&#39;], optional): Quantisation of the model. Defaults to 'fp16'.
        revision (Optional[str], optional): Revision of the repository if a HuggingFace repository name is given. Defaults to None.
        adapter_path (Optional[str], optional): Check the existence of the adapter file if given. Defaults to None.

    Returns:
        Path: Path to the local model to serve.
    """
    model_base_name = get_model_name(model_id_or_path=model_id_or_path)
    if quant != 'fp16':
        model_base_name += f'-{quant}'
    mlx_path = os.path.join(package_cache_dir, model_base_name)
    
    if os.path.exists(mlx_path):
        if adapter_path is not None:
            if not os.path.exists(adapter_path):
                raise FileNotFoundError(f'Adapter not found in "{adapter_path}".')
        return Path(mlx_path)
    else:
        hf_path = get_model_path(path_or_hf_repo=model_id_or_path, revision=revision)
        if quant == 'fp16':
            mlx_path = hf_path
        else:
            config = load_config(model_path=hf_path)
            quantisation = config.get('quantization', {}).get('bits')
            quant_target = int(quant.removeprefix('q'))
            if quant_target == quantisation:
                mlx_path = hf_path
            else:
                model_base_name = get_model_name(model_id_or_path=model_id_or_path)
                mlx_path = os.path.join(package_cache_dir, model_base_name + f'-q{quant_target}')
                if not os.path.exists(mlx_path):
                    convert(hf_path=hf_path, mlx_path=mlx_path, quantize=True, q_bits=int(quant[-1]))
        if adapter_path is not None:
            if not os.path.exists(adapter_path):
                raise FileNotFoundError(f'Adapter not found in "{adapter_path}".')
        return Path(mlx_path)
    
def get_model_and_tokenizer(
        model_id_or_path: str,
        tokenizer_id_or_path: Optional[str] = None,
        quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16',
        revision: Optional[str] = None,
        adapter_path: Optional[str] = None,
        model_config: Optional[Dict[str, Any]] = None,
        tokenizer_config: Optional[Dict[str, Any]] = None,
        logger: Optional[Logger] = None
    ) -> Tuple[nn.Module, TokenizerWrapper]:
    """Load the llm model and it's tokenizer.

    Args:
        model_id_or_path (str): HuggingFace repository name or the local path of the model.
        tokenizer_id_or_path (Optional[str], optional): HuggingFace repository name or the local path of the model. If None is given, `model_id_or_path` will be used. Defaults to None.
        quant (Literal[&#39;fp16&#39;, &#39;q8&#39;, &#39;q4&#39;, &#39;q2&#39;], optional): Quantisation of the model. Defaults to 'fp16'.
        revision (Optional[str], optional): Revision of the repository if a HuggingFace repository name is given. Defaults to None.
        adapter_path (Optional[str], optional): The adapter file for the model. Defaults to None.
        model_config (Optional[Dict[str, Any]], optional): Extra arguments for loading the model. Defaults to None.
        tokenizer_config (Optional[Dict[str, Any]], optional): Extra arguments for loading the tokenizer. Defaults to None.

    Returns:
        Tuple[nn.Module, TokenizerWrapper]: The model and the tokenizer.
    """
    mlx_path = make_model_exist(model_id_or_path=model_id_or_path, quant=quant, revision=revision, adapter_path=adapter_path)
    tokenizer_config = {} if tokenizer_config is None else tokenizer_config
    model_config = {} if model_config is None else model_config
    model = load_model(model_path=mlx_path, model_config=model_config)
    tokenizer = load_tokenizer(model_path=mlx_path if tokenizer_id_or_path is None else Path(tokenizer_id_or_path), tokenizer_config_extra=tokenizer_config, logger=logger)
    return model, tokenizer

class ModelConfig(NamedTuple):
    model_id_or_path: str
    tokenizer_id_or_path: Optional[str] = None
    adapter_path: Optional[str] = None
    quant: Literal['fp16', 'q8', 'q4', 'q2'] = 'fp16'
    revision: Optional[str] = None
    model_name: Optional[str] = None
    model_config: Optional[Dict[str, Any]] = None
    tokenizer_config: Optional[Dict[str, Any]] = None


from .cache_utils import CacheHistory, save_cache, load_cache, find_max_prefix_num, split_cache, trim_cache
from mlx_lm.models.cache import KVCache
import mlx.core as mx
import json, os
from logging import Logger
from copy import deepcopy
from datetime import datetime as dt
from typing import Optional, List, Union, Dict, NamedTuple, Tuple

class PromptSummary(NamedTuple):
    token_ids: List[int]
    length: int
    last_modified: dt

def prompt_summary_to_dict(ps: PromptSummary) -> Dict[str, Union[List[int], float]]:
    return dict(token_ids=ps.token_ids, last_modified=ps.last_modified.timestamp())

class CacheManager:
    """Class to manage multiple cache histories.
    """
    def __init__(self, cache_dir: str, token_threshold: int = 20, save_similarity_threshold: float = 0.9, max_keep: int = 50, logger: Optional[Logger] = None) -> None:
        """Initialise the cache manager.

        Args:
            cache_dir (str): The directory to save all the caches.
            token_threshold (int, optional): Minimum number of tokens to be considered to save as a prompt cache history. Defaults to 50.
            max_keep (int, optional): Maximum number of cache history files to keep. Defaults to 10.
            logger (Optional[Logger], optional): If a logger is provided, messages of adding and deleting caches will be added.
        """
        self.cache_dir = os.path.abspath(cache_dir)
        self.json_dir = os.path.join(self.cache_dir, 'cached_prompts.json')
        self.token_threshold = token_threshold
        self.save_sim_threshold = save_similarity_threshold
        self.max_keep = max_keep
        self.logger = logger
        if not os.path.exists(self.json_dir):
            os.makedirs(self.cache_dir, exist_ok=True)
            prompts = dict()
            with open(self.json_dir, 'w') as f:
                json.dump(prompts, f)

    @property
    def cached_prompts(self) -> Dict[str, PromptSummary]:
        """Dictionary that contains the information of all the currently stored cache histories.

        Returns:
            Dict[str, PromptSummary]: Dictionary that contains the information of all the currently stored cache histories.
        """
        with open(self.json_dir, 'r') as f:
            prompts = json.load(f)
        for k, v in prompts.items():
            v['last_modified'] = dt.fromtimestamp(v['last_modified'])
            v['length'] = len(v['token_ids'])
            prompts[k] = PromptSummary(**v)   
        return prompts
    
    @property
    def new_cache_id(self) -> str:
        """An unused cache ID for a new cache history.

        Returns:
            str: An unused cache ID for a new cache history.
        """
        current_ids: List[str] = list(self.cached_prompts.keys())
        current_ids = [int(x.removeprefix('cache_')) for x in current_ids]
        if len(current_ids) == 0:
            return 'cache_0'
        max_id = max(current_ids)
        for i in range(max_id):
            if i not in current_ids:
                return f'cache_{i}'
        return f'cache_{max_id + 1}'
    
    def _get_new_cache_id(self, exist_ids: List[str]) -> str:
        current_ids = [int(x.removeprefix('cache_')) for x in exist_ids]
        if len(current_ids) == 0:
            return 'cache_0'
        max_id = max(current_ids)
        for i in range(max_id):
            if i not in current_ids:
                return f'cache_{i}'
        return f'cache_{max_id + 1}'

    
    def _log_info(self, message: str) -> None:
        if self.logger is not None:
            self.logger.info(message)
    
    def drop_cache(self, cache_id: Union[str, List[str]]) -> None:
        """Drop an existing cache history.

        Args:
            cache_id (Union[str, List[str]]): The cache ID or list of cache IDs to drop.
        """
        cached_prompts = self.cached_prompts
        cache_id = [cache_id] if isinstance(cache_id, str) else list(set(cache_id))
        for cid in cache_id:
            if cid in cached_prompts.keys():
                ps = cached_prompts.pop(cid)
                cache_file = os.path.join(self.cache_dir, f'{cid}.safetensors')
                if os.path.exists(cache_file):
                    os.remove(cache_file)
                    self._log_info(f'Cache history "{cid}" dropped. Number of tokens: {ps.length}.')
        for k, v in cached_prompts.items():
            cached_prompts[k] = prompt_summary_to_dict(v)
        with open(self.json_dir, 'w') as f:
            json.dump(cached_prompts, f, indent=4)

    def find_cache(self, token_ids: Union[List[int], List[List[int]]], token_threshold: Optional[int] = None) -> Tuple[Optional[CacheHistory], Optional[List[str]]]:
        """Find a cache history given the new prompt token ids.

        Args:
            token_ids (List[int]): The new prompt token ids.
            token_threshold (Optional[int], optional): Minimum number of tokens in the cache history to be considered to use. If None is given, token_threshold of the cache manager will be used as default. Defaults to None.

        Returns:
            Tuple[Optional[CacheHistory], Optional[str]]: The cache history and it's cache ID.
        """
        # Make them list of list of tokens
        token_ids = [token_ids] if isinstance(token_ids[0], int) else token_ids
        prompt_lens = [len(tids) for tids in token_ids]
        max_prompt_len = max(prompt_lens)

        # Calculate the total number of prompts
        num_prompts = len(token_ids)
        self._log_info(f'Number of prompts: {num_prompts}')

        # Make the prompt tokens tuples to find unique sets for querying, avoiding repetitive search for large batches
        token_tuples = [tuple(tids) for tids in token_ids]
        prompt_dict = dict(zip(range(num_prompts), token_tuples))
        find_queries = list(set(token_tuples))
        
        # Get existing prompt caches for search
        cached_prompts = self.cached_prompts

        # Early exit if no existing prompts
        if len(cached_prompts) == 0:
            self._log_info('No existing cache.')
            return None, None
        # Get the number of token token threshold
        token_threshold = self.token_threshold if token_threshold is None else token_threshold

        # Searching query by query
        found_cache = dict()
        for query in find_queries:
            q_len = len(query)
            res = [(
                k, # cache id
                find_max_prefix_num(query, v.token_ids), # number of tokens to get
                q_len - len(v.token_ids), # inverse number of tokens in the cache, easier to sort
                max_prompt_len - q_len, # prompt offset due to padding
                ) for k, v in cached_prompts.items()] # Get the highest number of shared prefix with smallest cache
            res.sort(key=lambda x: x[1:3], reverse=True)
            found_cache[query] = res[0] if res[0][1] >= token_threshold else (None, 0, 0, 0)

        prompt_dict = {k: found_cache[v] for k, v in prompt_dict.items()}
        used_cache = [v[0] for v in found_cache.values() if v[0] is not None]
        used_cache = list(set(used_cache))

        if len(used_cache) == 0:
            self._log_info(f'No cache with prefix match tokens larger than {token_threshold}.')
            return None, None

        # Load the caches
        cache_list = []
        for cid in used_cache:
            cache, metadata = load_cache(os.path.join(self.cache_dir, f'{cid}.safetensors'))
            cache_list.append((cid, cache))

        # Calculate the max position of the cache
        max_cache_len = min([v[3] + v[1] for v in found_cache.values()])
        max_cache_len = max_cache_len - 1 if max_cache_len == max_prompt_len else max_cache_len

        # Making empty keys and values pairs of tensors from found cache
        sample_cache = cache_list[0][1]
        kv_list = []
        for layer in sample_cache:
            _n, num_kv_heads, _nt, head_dim = layer.state[0].shape
            kv_list.append(
                (
                    mx.zeros((num_prompts, num_kv_heads, max_cache_len, head_dim), dtype=layer.state[0].dtype),
                    mx.zeros((num_prompts, num_kv_heads, max_cache_len, head_dim), dtype=layer.state[1].dtype)
                )
            )
        cache_dict = dict(cache_list)
        
        # filling the tensors with old cache
        new_cache = []
        for i, (keys, values) in enumerate(kv_list):
            new_cache.append(KVCache())
            for p_index, p_ids in prompt_dict.items():
                cid, max_prefix, _, offset = p_ids
                if max_cache_len - offset > 0:
                    keys[p_index:p_index + 1, :, offset:, :] = cache_dict[cid][i].state[0][:, :, :max_cache_len - offset, :]
                    values[p_index:p_index + 1, :, offset:, :] = cache_dict[cid][i].state[1][:, :, :max_cache_len - offset, :]      
            new_cache[i].update_and_fetch(keys, values)
            mx.eval(new_cache[i].state)
        self.update_cache_time(used_cache)
        uc_str = '", "'.join(used_cache)
        self._log_info(f'Using existing cache "{uc_str}" for {max_cache_len} tokens for each prompt.')
        del cache_list, cache_dict
        mx.metal.clear_cache()
        return CacheHistory(cache=new_cache, token_ids=[]), used_cache

    def update_cache_time(self, cache_id: Union[str, List[str]]) -> None:
        """Updating the last modified time of an existing cache history.

        Args:
            cache_id (str): The existing cache ID to update.
        """
        cache_id = [cache_id] if isinstance(cache_id, str) else cache_id
        cached_prompts = self.cached_prompts
        for k, v in cached_prompts.items():
            cached_prompts[k] = prompt_summary_to_dict(v)
            if k in cache_id:
                cached_prompts[k]['last_modified'] = dt.now().timestamp()
        with open(self.json_dir, 'w') as f:
            json.dump(cached_prompts, f, indent=4)

    def save_cache(self, 
        cache: CacheHistory,
        token_threshold: Optional[int] = None, 
    ) -> None:
        """Save a new cache history.

        Args:
            cache (Union[StepOutput, CacheHistory]): The new cache history to save.
            replace_threshold (float, optional): Percentage threshold of tokens prefix similar to the given from_cache_id to replace the given cache ID. Defaults to 0.9.
            token_threshold (Optional[int], optional): Minimum number of tokens in the cache history to be considered to be saved. If None is given, token_threshold of the cache manager will be used as default. Defaults to None.
            from_cache_id (Optional[str], optional): A cache ID that will be considered to be replaced (see replace_threshold). Defaults to None.
            metadata (Optional[Dict[str, str]], optional): Metadata to save in the cache history file. Defaults to None.
        """
        token_threshold = self.token_threshold if token_threshold is None else token_threshold
        save_threshold = self.save_sim_threshold
        token_ids = [tuple(tids) for tids in cache.token_ids]
        prompt_lens = [len(tids) for tids in cache.token_ids]
        max_lens = max(prompt_lens)
        offsets = [max_lens - pl for pl in prompt_lens]
        caches = split_cache(cache=cache.cache)
        unique_caches_tokens = []
        unique_caches = []
        for i, tids in enumerate(token_ids):
            if (tids not in unique_caches_tokens) and prompt_lens[i] >= token_threshold:
                unique_caches_tokens.append(tids)
                unique_caches.append((tids, prompt_lens[i], offsets[i], trim_cache(caches[i], offset=offsets[i])))
        del caches, cache
        mx.metal.clear_cache()
        if len(unique_caches) == 0:
            self._log_info(f'No new caches saved as they are shorter than {token_threshold} tokens.')
            return
        
        # Compare among the unique caches and see if they have similar ones
        to_drop = []
        for tids, plen, offset, cache in unique_caches:
            if tids in unique_caches_tokens:
                comp_list = list(filter(lambda x: x[0] != tids, unique_caches))
                for comp in comp_list:
                    max_pref = find_max_prefix_num(tids, comp[0])
                    osim = max_pref / plen
                    csim = max_pref / comp[1]
                    max_sim = max(osim, csim)
                    if max_sim > save_threshold:
                        drop_tids = tids if osim == max_sim else comp[0]
                        to_drop.append(drop_tids)
        unique_caches = list(filter(lambda x: x[0] not in to_drop, unique_caches))

        # Comparing with existing caches
        exist_to_drop = []
        exist_to_keep = []
        cached_prompts = self.cached_prompts
        for tids, plen, offset, cache in unique_caches:
            if tids in unique_caches_tokens:
                for cid, ps in cached_prompts.items():
                    max_pref = find_max_prefix_num(tids, ps.token_ids)
                    osim = max_pref / plen
                    csim = max_pref / ps.length
                    max_sim = max(osim, csim)
                    if max_sim > save_threshold:
                        if osim == max_sim:
                            to_drop.append(tids)
                            exist_to_keep.append(cid)
                        else:
                            exist_to_drop.append(cid)
        unique_caches = list(filter(lambda x: x[0] not in to_drop, unique_caches)) 
        self.drop_cache(exist_to_drop)   

        # Final number of prompts to save
        num_save = len(unique_caches)
        if num_save > 0:
            cached_prompts = self.cached_prompts
            exist_to_drop = None
            if len(cached_prompts) + num_save > self.max_keep:
                num_to_drop = len(cached_prompts) + num_save - self.max_keep
                cp_ordered = list(cached_prompts.items())
                cp_ordered.sort(key=lambda x: x[1].last_modified)
                cp_ordered = list(filter(lambda x: x[0] not in exist_to_keep, cp_ordered))
                exist_to_drop = [x[0] for x in cp_ordered[:num_to_drop]]
            if exist_to_drop:
                self.drop_cache(exist_to_drop)
            lm = dt.now()
            cached_prompts = self.cached_prompts
            exist_ids = list(cached_prompts.keys())
            for tids, length, offset, che in unique_caches:
                new_id = self._get_new_cache_id(exist_ids=exist_ids)
                cached_prompts[new_id] = PromptSummary(token_ids=list(tids), length=length, last_modified=lm)
                exist_ids.append(new_id)
                save_cache(che, file=os.path.join(self.cache_dir, f'{new_id}.safetensors'))
                self._log_info(f'New cache "{new_id}" saved. Prompt length: {length}.')
            cached_prompts = {k: prompt_summary_to_dict(v) for k, v in cached_prompts.items()}
            with open(self.json_dir, 'w') as f:
                json.dump(cached_prompts, f, indent=4)
        else:
            self._log_info(f'New caches are not saved as old ones covered them already.')
        return



# adapted from mlx-lm
import json
from functools import partial
from typing import List, Optional, Dict, Any
from logging import Logger
from transformers import AutoTokenizer, PreTrainedTokenizer
from huggingface_hub import hf_hub_download

REPLACEMENT_CHAR = "\ufffd"
SPECIAL_SPACE = "\u2581"

def _remove_space(x):
    if x and x[0] == " ":
        return x[1:]
    return x


class NaiveDetokenizer:

    def __init__(self, tokenizer: PreTrainedTokenizer) -> None:
        self._tokenizer = tokenizer
        self.reset()

    def reset(self, num_seqs: Optional[int] = None) -> None:
        self.num_seqs = num_seqs
        self.texts = [] if self.num_seqs is None else [''] * self.num_seqs
        self.tokens = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.current_tokens = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.final_segments = None

    def add_tokens(self, token_ids: List[List[int]]) -> None:
        if self.num_seqs is None:
            self.reset(num_seqs=len(token_ids))
        elif len(token_ids) != self.num_seqs:
            raise Exception('Number of tokens does not match with number of existing sequences.')
        self.current_tokens = [tids + nids for tids, nids in zip(self.current_tokens, token_ids)]

    def finalize(self) -> None:
        new_texts = self._tokenizer.batch_decode(self.current_tokens)
        new_texts = [nt.rstrip(REPLACEMENT_CHAR) for nt in new_texts]
        self.tokens = [t + ct for t, ct in zip(self.tokens, self.current_tokens)]
        self.texts = [t + nt for t, nt in zip(self.texts, new_texts)]
        self.current_tokens = [[]] * self.num_seqs
        self.final_segments = new_texts        

    @property
    def last_segments(self) -> List[str]:
        if self.final_segments is not None:
            return self.final_segments
        new_texts = self._tokenizer.batch_decode(self.current_tokens)
        with_repl = [newt.endswith(REPLACEMENT_CHAR) for newt in new_texts]
        bundle = [(ot if wr else ot + nt, nt if wr else [], os if wr else os + ns, '' if wr else ns) for ot, nt, os, ns, wr in zip(self.tokens, self.current_tokens, self.texts, new_texts, with_repl)]
        tokens, current_tokens, texts, new_segments = list(zip(*bundle))
        self.tokens = list(tokens)
        self.current_tokens = list(current_tokens)
        self.texts = list(texts)
        new_segments = list(new_segments)
        return new_segments

class SPMDetokenizer:
    """A streaming detokenizer for SPM models.

    It adds tokens to the text if the next token starts with the special SPM
    underscore which results in linear complexity.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, trim_space=True):
        self.trim_space = trim_space
        self.eos_token = tokenizer.eos_token

        # Extract the tokens in a list from id to text
        self.tokenmap = [""] * (max(tokenizer.vocab.values()) + 1)
        for value, tokenid in tokenizer.vocab.items():
            self.tokenmap[tokenid] = value

        self.hexcode_tokens = [i for i, t in enumerate(self.tokenmap) if t.startswith('<0x')]

        self.reset()

    def reset(self, num_seqs: Optional[int] = None) -> None:
        self.num_seqs = num_seqs
        self.range = [] if self.num_seqs is None else list(range(self.num_seqs))
        self.texts = [] if self.num_seqs is None else [''] * self.num_seqs
        self.tokens = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.hexcodes = [] if self.num_seqs is None else [[]] * self.num_seqs
        self.segments = [] if self.num_seqs is None else [''] * self.num_seqs

    def _get_text_token(self, token_id, raw_token, token_condition, index) -> str:
        self.tokens[index].append(token_id[0])
        is_space, is_hex = token_condition
        output = ''
        if is_hex:
            self.hexcodes[index].append(int(raw_token[3:5], 16))
        elif is_space:
            if self.texts[index] or (not self.trim_space):
                output = ('' if len(self.hexcodes[index]) == 0 else bytes(self.hexcodes[index]).decode()) + raw_token.replace(SPECIAL_SPACE, ' ')
            else:
                output = ('' if len(self.hexcodes[index]) == 0 else bytes(self.hexcodes[index]).decode()) + _remove_space(raw_token.replace(SPECIAL_SPACE, ' '))
            self.hexcodes[index] = []
        else:
            output = ('' if len(self.hexcodes[index]) == 0 else bytes(self.hexcodes[index]).decode()) + raw_token
            self.hexcodes[index] = []
        self.texts[index] += output
        return output

    def add_tokens(self, token_ids: List[List[int]]) -> None:
        if self.num_seqs is None:
            self.reset(num_seqs=len(token_ids))
        elif len(token_ids) != self.num_seqs:
            raise Exception('Number of tokens does not match with number of existing sequences.')
        raw_tokens = [self.tokenmap[token[0]] for token in token_ids]
        token_conditions = [((rt[0] == SPECIAL_SPACE), (tid[0] in self.hexcode_tokens)) for rt, tid in zip(raw_tokens, token_ids)]
        self.segments = [self._get_text_token(tid, rt, tc, i) for tid, rt, tc, i in zip(token_ids, raw_tokens, token_conditions, self.range)]

    def finalize(self):
        hex_str = [('' if len(self.hexcodes[index]) == 0 else bytes(self.hexcodes[index]).decode()) for index in self.range]
        self.segments = [s + hs for s, hs in zip(self.segments, hex_str)]

    @property
    def last_segments(self):
        """Return the last segment of readable text since last time this property was accessed."""
        segments = self.segments
        self.segments = [''] * self.num_seqs
        return segments


class TokenizerWrapper:
    """A wrapper that combines an HF tokenizer and a detokenizer.

    Accessing any attribute other than the ``detokenizer`` is forwarded to the
    huggingface tokenizer.
    """

    def __init__(self, tokenizer, detokenizer_class=NaiveDetokenizer):
        self._tokenizer = tokenizer
        self._detokenizer_class = detokenizer_class
        self._detokenizer = detokenizer_class(tokenizer)

    def __getattr__(self, attr):
        if attr == "detokenizer":
            return self._detokenizer
        elif attr.startswith("_"):
            return self.__getattribute__(attr)
        else:
            return getattr(self._tokenizer, attr)

    def __setattr__(self, attr, value):
        if attr == "detokenizer":
            raise AttributeError("Cannot set the detokenizer.")
        elif attr.startswith("_"):
            super().__setattr__(attr, value)
        else:
            setattr(self._tokenizer, attr, value)



def _match(a, b):
    if type(a) != type(b):
        return False
    if isinstance(a, dict):
        return len(a) == len(b) and all(k in b and _match(a[k], b[k]) for k in a)
    if isinstance(a, list):
        return len(a) == len(b) and all(_match(ai, bi) for ai, bi in zip(a, b))

    return a == b


def _is_spm_decoder(decoder):
    _target_description = {
        "type": "Sequence",
        "decoders": [
            {"type": "Replace", "pattern": {"String": "▁"}, "content": " "},
            {"type": "ByteFallback"},
            {"type": "Fuse"},
            {"type": "Strip", "content": " ", "start": 1, "stop": 0},
        ],
    }
    return _match(_target_description, decoder)


def _is_spm_decoder_no_space(decoder):
    _target_description = {
        "type": "Sequence",
        "decoders": [
            {"type": "Replace", "pattern": {"String": "▁"}, "content": " "},
            {"type": "ByteFallback"},
            {"type": "Fuse"},
        ],
    }
    return _match(_target_description, decoder)


def _is_bpe_decoder(decoder):
    return isinstance(decoder, dict) and decoder.get("type", None) == "ByteLevel"


def load_tokenizer(model_path, tokenizer_config_extra: Optional[Dict[str, Any]] = None, logger: Optional[Logger] = None):
    """Load a huggingface tokenizer and try to infer the type of streaming
    detokenizer to use.

    Note, to use a fast streaming tokenizer, pass a local file path rather than
    a Hugging Face repo ID.
    """
    tokenizer_config_extra = dict() if tokenizer_config_extra is None else tokenizer_config_extra
    detokenizer_class = NaiveDetokenizer

    tokenizer_file = model_path / "tokenizer.json"
    if not tokenizer_file.exists():
        tokenizer_file = hf_hub_download(repo_id=str(model_path), filename='tokenizer.json')
    with open(tokenizer_file, "r") as fid:
        tokenizer_content = json.load(fid)
    if "decoder" in tokenizer_content:
        if _is_spm_decoder(tokenizer_content["decoder"]):
            if logger:
                logger.info('Using SPM decoder.')
            detokenizer_class = SPMDetokenizer
        elif _is_spm_decoder_no_space(tokenizer_content["decoder"]):
            if logger:
                logger.info('Using SPM decoder with trim_space=False.')
            detokenizer_class = partial(SPMDetokenizer, trim_space=False)
        elif logger:
            logger.info('Using Naive decoder.')

    return TokenizerWrapper(
        AutoTokenizer.from_pretrained(model_path, **tokenizer_config_extra),
        detokenizer_class,
    )


from mlx_lm.tokenizer_utils import TokenizerWrapper
from pydantic import BaseModel
from transformers import PreTrainedTokenizerBase
import json
from typing import List, Dict, Optional, Any, Literal, Union

def get_chat_prompt(msgs: List[Dict[str, str]], tokenizer: TokenizerWrapper) -> str:
    last_role = msgs[-1]['role']
    if last_role != 'assistant':
        return tokenizer.apply_chat_template(msgs, add_generation_prompt=True, tokenize=False)
    else:
        last_content = msgs[-1]['content']
        lc_strip = last_content.strip()
        wa = tokenizer.apply_chat_template(msgs, add_generation_prompt=True, tokenize=False)
        woa = tokenizer.apply_chat_template(msgs[:-1], add_generation_prompt=True, tokenize=False)
        diff = wa.removeprefix(woa)
        last_chunk = lc_strip
        if last_content in diff:
            last_chunk = last_content
        pref = ''.join(diff.split(last_chunk)[:-1]) + last_content.lstrip()
        return woa + pref
    
def convert_tool_to_json_schema(tool: Dict[str, Any]) -> Dict[str, Any]:
    params_json = tool.get('function', dict()).get('parameters')
    if params_json is not None:
        if 'title' not in params_json.keys():
            params_json['title'] = tool['function']['name']
        return params_json
    else:
        return tool
    
def get_tool_name(tool: Dict[str, Any]) -> str:
    # OpenAI format
    name = tool.get('function', dict()).get('name')
    if name is None:
        name = tool.get('title')
    return name
    
TOOL_LIST = [{'properties': {'example_arg': {'title': 'Example Arg', 'type': 'string'}}, 'required': ['example_arg'], 'title': 'ExampleTool', 'type': 'function'}]
TOOL_CALL_LIST = [dict(function=dict(name='ExampleTool', arguments=dict(example_arg='example')))]
MSG_WITH_TOOL = [
    dict(role='user', content='Hello'),
    dict(role='assistant', tool_calls=TOOL_CALL_LIST)
]
MSG_SINGLE_ASSISTANT = [dict(role='assistant', content='')]
DEFAULT_TOOL_SYSTEM = '''

# Tools

You may call one or more functions to assist with the user query.

You are provided with function signatures within <tools></tools> XML tags:
<tools>
$$tool_list$$
</tools>

For each function call, return a json object with function name and arguments within <tool_call></tool_call> XML tags:
<tool_call>
{"name": <function-name>, "arguments": <args-json-object>}
</tool_call>'''

class ToolCallContent(BaseModel):
    name: str
    arguments: str

class ToolCall(BaseModel):
    function: ToolCallContent

class ChatMessage(BaseModel):
    role: Literal['system', 'user', 'assistant', 'tool']
    content: Optional[Any] = None
    tool_calls: Optional[List[ToolCall]] = None


class ChatTemplate:
    def __init__(self, tokenizer: PreTrainedTokenizerBase) -> None:
        self.tokenizer = tokenizer

    @property
    def support_tool_call(self) -> bool:
        if not hasattr(self, '_support_tool_call'):
            try:
                p_with_tool = self.tokenizer.apply_chat_template(MSG_WITH_TOOL[:1], tools=TOOL_LIST, add_generation_prompt=True, tokenize=False)
                p_wo_tool = self.tokenizer.apply_chat_template(MSG_WITH_TOOL[:1], add_generation_prompt=True, tokenize=False)
                self._support_tool_call = p_with_tool != p_wo_tool
            except:
                self._support_tool_call = False
        return self._support_tool_call
    
    @property
    def support_system(self) -> bool:
        if not hasattr(self, '_support_system'):
            try:
                system = 'Test system message, see if exist.'
                messages = [dict(role='system', content=system), dict(role='user', content='Hi there')]
                prompt = self.tokenizer.apply_chat_template(messages, add_generation_prompt=True, tokenize=False)
                self._support_system = system in prompt
            except:
                self._support_system = False
        return self._support_system

    @property
    def allow_multiple_assistant(self) -> bool:
        if not hasattr(self, '_allow_multiple_assistant'):
            try:
                self.tokenizer.apply_chat_template(MSG_SINGLE_ASSISTANT * 2, tokenize=False, add_generation_prompt=False, continue_final_message=True)
                self._allow_multiple_assistant = True
            except:
                self._allow_multiple_assistant = False
        return self._allow_multiple_assistant

    @property
    def tool_start(self) -> str:
        if not hasattr(self, '_tool_start'):
            if self.support_tool_call:
                p_with_tool = self.tokenizer.apply_chat_template(MSG_WITH_TOOL, tools=TOOL_LIST, add_generation_prompt=True, tokenize=False)
                p_wo_tool = self.tokenizer.apply_chat_template(MSG_WITH_TOOL[:1], tools=TOOL_LIST, add_generation_prompt=False, tokenize=False)
                diff_str = p_with_tool.removeprefix(p_wo_tool)
                tool_first_index = diff_str.find('{')
                self._tool_start = diff_str[:tool_first_index]
            else:
                self._tool_start = '<tool_call>\n'
        return self._tool_start
    
    def _validate_messages(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        for msg in messages:
            ChatMessage(**msg)
        if len(messages) == 0:
            raise Exception(f'Must have at least one message.')
        return messages
    
    def _validate_message_seq(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        def format_tool_call(tc):
            call = tc['function']
            try:
                args = json.loads(call['arguments'])
            except:
                args = call['arguments']
            call['arguments'] = args
            return json.dumps(call)
        if self.allow_multiple_assistant and self.support_tool_call:
            new_messages = []
            for msg in messages:
                if msg.get('tool_calls') is not None:
                    msg['tool_calls'] = [format_tool_call(tc) for tc in msg['tool_calls']]
                    new_messages.append(msg)
                else:
                    new_messages.append(msg)
        elif self.allow_multiple_assistant and (not self.support_tool_call):
            new_messages = []
            for msg in messages:
                if (msg['role'] == 'assistant') and msg.get('tool_calls'):
                    tool_call = '\n'.join(['<tool_call>\n' + format_tool_call(tc) + '\n</tool_call>' for tc in msg['tool_calls']])
                    content = '' if msg.get('content') is None else msg.get('content')
                    new_messages.append(dict(role='assistant', content=content + tool_call))
                elif (msg['role'] == 'tool'):
                    content = '' if msg.get('content') is None else msg.get('content')
                    content = content if isinstance(content, str) else json.dumps(content)
                    content = '<tool_response>\n' + content + '\n</tool_response>'
                    new_messages.append(dict(role='user', content=content))
                else:
                    new_messages.append(msg)
        elif self.support_tool_call:
            new_messages = []
            last_role = None
            for msg in messages:
                if (msg['role'] == 'assistant') and last_role != 'user':
                    if msg.get('tool_calls') is not None:
                        msg['tool_calls'] = [format_tool_call(tc) for tc in msg['tool_calls']]
                    new_messages.extend([dict(role='user', content=''), msg])
                else:
                    new_messages.append(msg)
                last_role = msg['role']
        else:
            new_messages = []
            last_role = None
            for msg in messages:
                if (msg['role'] == 'assistant') and msg.get('tool_calls'):
                    tool_call = '\n'.join(['<tool_call>\n' + format_tool_call(tc) + '\n</tool_call>' for tc in msg['tool_calls']])
                    content = '' if msg.get('content') is None else msg.get('content')
                    to_append = dict(role='assistant', content=content + tool_call)
                elif (msg['role'] == 'tool'):
                    content = '' if msg.get('content') is None else msg.get('content')
                    content = content if isinstance(content, str) else json.dumps(content)
                    content = '<tool_response>\n' + content + '\n</tool_response>'
                    to_append = dict(role='user', content=content)
                else:
                    to_append = msg
                if (to_append['role'] == 'assistant') and last_role != 'user':
                    new_messages.extend([dict(role='user', content=''), to_append])
                else:
                    new_messages.append(to_append)
                last_role = to_append['role']
        if not self.support_system and (new_messages[0]['role'] == 'system'):
            system = new_messages[0]['content']
            new_messages = new_messages[1:]
            if (len(new_messages) > 0) and (new_messages[0]['role'] == 'user'):
                first_message = new_messages[0]['content']
                first_message = '<system>\n' + system + '\n</system>\n\n' + first_message
                new_messages[0]['content'] = first_message
            else:
                raise Exception(f'First message after the system message is not a user message.')
        return new_messages            

    def apply_chat_template(self, 
            messages: List[Dict[str, Any]], 
            tools: Optional[List[Dict[str, Any]]] = None, 
            tool_choice: Union[Literal['none', 'auto', 'required'], Dict[str, Union[str, Dict[str, str]]]] = 'auto',
            add_generation_prompt: bool = True
            ) -> str:
        messages = self._validate_messages(messages=messages)
        messages = self._validate_message_seq(messages=messages)
        tools = tools if tool_choice != 'none' else None
        continue_final_message = False
        if messages[-1]['role'] =='assistant':
            continue_final_message = True
            add_generation_prompt = False
        if tools and (not self.support_tool_call):
            tool_json = '\n'.join([json.dumps(tool) for tool in tools])
            if messages[0]['role'] == 'system':
                content = messages[0].get('content', '') + DEFAULT_TOOL_SYSTEM.replace('$$tool_list$$', tool_json)
                messages[0]['content'] = content.strip()
            else:
                messages = [dict(role='system', content=DEFAULT_TOOL_SYSTEM.replace('$$tool_list$$', tool_json).strip())] + messages
        prompt = self.tokenizer.apply_chat_template(conversation=messages, 
            tools=tools if self.support_tool_call else None, tokenize=False, 
            add_generation_prompt=add_generation_prompt, 
            continue_final_message=continue_final_message)
        if tool_choice == 'required':
            prompt += self.tool_start + '{"name": "'
        elif isinstance(tool_choice, dict):
            tool_name = tool_choice['function']['name']
            prompt += self.tool_start + '{' + f'"name": "{tool_name}", arguments": '
        return prompt


